/*
 * beam.c
 * - code to beam texts across the TCP connection following a
 * special protocol
 * source file for powwow
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "powwow.h"
#include "beam.h"
#include "mapping.h"
#include "tcp.h"
#include "tty.h"
#include "eval.h"

static void finish_edit();

editsess *edit_sess;	/* head of session list */

char edit_start[BUFSIZE]; /* messages to send to host when starting */ 
char edit_end[BUFSIZE];   /* or leaving editing sessions */ 

void write_message(s)
char *s;
{
  tty_clear_input_line(opt_compact);
  if(!opt_compact) tty_printf("\n");
  tty_puts(s);
  /* col0 = promptlen = promptstr[0] = surely_isprompt = 0; */
  showprompt();
  tty_draw_input_line();
}
/*
 * Process editing protocol message from buf with len remaining chars.
 * Return number of characters used in the message.
 */
int process_message(buf, len)
char *buf; int len;
{
  int msglen, i, l, used;
  char *text;
  char msg[LINELEN];

  msglen = atoi(buf + 1);
  for(i = 1; i < len && isdigit(buf[i]); i++);
  if(buf[i] == '\r') i++;
  if(buf[i++] != '\0' || i >= len) {
     write_message ("#warning: MPI protocol error\n");
     return len;
  }
  l = len - i;
  if (msglen < l) {
    l = msglen;
    used = msglen + i;
  } else
    used = len;

  text = (char *)malloc(msglen);
  if (!text) {
    errmsg("malloc");
    return used > len ? len : used;
  }

  bcopy(buf + i, text, l);
  i = l;
  while(i < msglen) {
     /*
      * read() might block here, but it should't be long. I should also
      * process any telnet protocol commands, but I don't care right now.
      */
     while((l = read(tcp_fd, text + i, msglen - i)) == -1 && errno == EINTR);
     if (l == -1) {
     	errmsg("read message from socket");
	return used > len ? len : used;
     }
     tty_printf("\rgot %d bytes out of %d", i, msglen);
     tty_flush();
     i += l;
  }
  tty_printf("\rread all %d bytes.%s\n", msglen, tty_clreoln);
  switch(*buf) {
    case 'E':
     message_edit(text, msglen, 0, 0);
     break;
    case 'V':
     message_edit(text, msglen, 1, 0);
     break;
    default:
     sprintf(msg, "#Warning: received a funny message (0x%x)\n", *buf);
     write_message(msg);
     free(text);
     break;
  }
  return used;
}

/*
 * start an editing session: process the EDIT/VIEW message
 * if view == 1, text will be viewed, else edited
 */
void message_edit(text, msglen, view, builtin)
char *text;
int msglen;
char view, builtin;
{
    char tmpname[LINELEN], command_str[LINELEN], buf[LINELEN];
    int tmpfd, key;
    int i;
    int childpid;
    editsess *s;
    char *editor, *descr;
    char *args[4];
    int waitforeditor;
    args[0] = "/bin/sh"; args[1] = "-c";
    args[3] = 0;
  
    args[2] = command_str;

    if(view) {
     key = -1;
	i = 0;
    } else {
	if (text[0] != 'M') {
	    tty_printf("#Warning: protocol error (message_edit, no M)\n");
	    free(text);
	    return;
	}
	for (i = 1; i < msglen && isdigit(text[i]); i++);
	if (text[i++] != '\n' || i >= msglen) {
	    tty_printf("#Warning: protocol error (message_edit, no \\n)\n");
	    free(text);
	    return;
	}
	key = atoi(text + 1);
    }
    descr = text + i;
    while (i < msglen && text[i] != '\n') i++;
    if (i >= msglen) {
	tty_printf("#Warning: protocol error (message_edit, no desc)\n");
	free(text);
	return;
    }
    text[i++] = '\0';
      
    sprintf(tmpname, "/tmp/powwow.%d.%d%d", getpid(), key, abs(rand()) >> 8);
    if ((tmpfd = open(tmpname, O_WRONLY | O_CREAT, 0600)) < 0) {
	errmsg("create temp edit file");
	free(text);
	return;
    }
    if (fchmod(tmpfd, S_IRUSR | S_IWUSR) < 0) {
	errmsg("chmod temp edit file");
	free(text);
	close(tmpfd);
	return;
    }
    if(write(tmpfd, text + i, msglen - i) < msglen - i) {
	errmsg("write to temp edit file");
	free(text);
	close(tmpfd);
	return;
    }
    close(tmpfd);
  
    s = (editsess*)malloc(sizeof(editsess));
    if (!s) {
	errmsg("malloc");
	return;
    }
    s->ctime = time((time_t*)NULL);
    s->oldsize = msglen - i;
    s->key = key;
    s->builtin = builtin;
    s->descr = str_dup(descr);
    s->file = str_dup(tmpname);
    free(text);
      
    /* send a edit_start message (if wanted) */ 
    if ((!edit_sess) && (*edit_start)) {
	error = 0;
	parse_instruction(edit_start, 0, 0);
	history_done = 0;
    }

    if (view) {
	if (!(editor = getenv("POWWOWPAGER")) && !(editor = getenv("PAGER")))
	  editor = "more";
    } else {
	if (!(editor = getenv("POWWOWEDITOR")) && !(editor = getenv("EDITOR")))
	  editor = "emacs";
    }
  
    if (editor[0] == '&') {
	waitforeditor = 0;
	editor++;
    } else
      waitforeditor = 1;
  
    if(waitforeditor) {
	tty_quit();
	/* ignore SIGINT since interrupting the child would interrupt us too,
	 if we are in the same tty group */
	sig_permanent(SIGINT, SIG_IGN);
	sig_permanent(SIGCHLD, SIG_DFL);
    }
  
    switch(childpid = fork()) {		/* let's get schizophrenic */
     case 0:
	sprintf (command_str, "%s %s", editor, s->file);
	sprintf (buf, "TITLE=%s", s->descr);
	putenv (buf);
	/*	   setenv("TITLE", s->descr, 1);*/
	execvp(args[0], args);
	syserr("execve");
	break;
     case -1:
	errmsg("fork");
	free(s->descr);
	free(s->file);
	free(s);
	return;
    }
    s->pid = childpid;
    if (waitforeditor) {
	while ((i = waitpid(childpid, (int*)NULL, 0)) == -1 && errno == EINTR)
	  ;
	signal_init();		/* reset SIGINT and SIGCHLD handlers */
	if (i == -1)
	  errmsg("waitpid");
	else
	  finish_edit(s);
	free(s->descr);
	free(s->file);
	tty_init();
	if (i == -1) {
	    free(s);
	    return;
	}
	if (!edit_sess && *edit_end) {
	    error = 0;
	    parse_instruction(edit_end, 0, 0);
	    history_done = 0;
	}

	if (!s->builtin) {
	    tty_gotoxy(0, lines - 1);
	    tty_printf("\n%s", promptstr);
	    tty_draw_input_line();
	    line0 = lines - 1;
	}
	free(s);

    } else {
	s->next = edit_sess;
	edit_sess = s;
    }
}

/*
 * Our child has snuffed it. check if it was an editor, and update the
 * session list if that is the case.
 */

void childdeathsig()
{
    int pid, ret;
    editsess **sp, *p;
   
    /* GH: while() instead of just one check */
    while ((pid = waitpid(-1, &ret, WNOHANG)) > 0) {

	/* GH: check for WIFSTOPPED unnescessary since no	*/
	/* WUNTRACED to waitpid()				*/
	for (sp = &edit_sess; *sp && (*sp)->pid != pid; sp = &(*sp)->next)
	  ;
	if (*sp) {
	    if ((*sp)->key >= 0)	/* send back reply message? */
		finish_edit(*sp);
	    p = *sp; *sp = p->next;
	    free(p->descr);
	    free(p->file);
	    free(p);

	    /* GH: only send message if found matching session */

	    /* send the edit_end message if this is the last editor... */ 
	    if ((!edit_sess) && (*edit_end)) {
		int otcp_fd = tcp_fd;  /* backup current socket fd */
		tcp_fd = tcp_main_fd;
		error = 0;
		parse_instruction(edit_end, 0, 0);  /* KLUDGE!! (oh well..) */
		history_done = 0;
		tcp_fd = otcp_fd;
	    }
	}
    }
}

/*
 * send back edited text to server, or cancel the editing session if the
 * file was not changed.
 */
static void finish_edit(sp)
editsess *sp;
{
    int fd;
    char *msg = NULL, *text = NULL;
    int txtlen, hdrlen;
    struct stat sbuf;
    char keystr[16];
    char buf[LINELEN];
    
    if (sp->builtin) {
	if (unlink(sp->file) < 0)
	  errmsg("unlink edit file");
	return;
    }
    fd = open(sp->file, O_RDONLY);
    if (fd == -1) {
	errmsg("open edit file");
	goto cleanup_file;
    }
    if (fstat(fd, &sbuf) == -1) {
	errmsg("fstat edit file");
	goto cleanup_fd;
    }
    
    txtlen = sbuf.st_size;
  
    if (sbuf.st_mtime > sp->ctime || txtlen != sp->oldsize) {
	/* file was changed by editor: send back result to server */
     
	text = (char*)malloc(txtlen + 1); /* +1 is for possible LF */
	if (!text) {
	    errmsg("malloc");
	    goto cleanup_fd;
	}
	if (read(fd, text, txtlen) < txtlen) {
	    errmsg("read edit file");
	    goto cleanup_text;
	}
	if (txtlen && text[txtlen - 1] != '\n') {
	    /* If the last line isn't LF-terminated, add an LF */
	    text[txtlen] = '\n';
	    txtlen++;
	}

	sprintf(keystr, "E%d\n", sp->key);
	msg = (char*)malloc(txtlen + 64);	/* 64 is safety margin */
	if (!msg) {
	    errmsg("malloc");
	    goto cleanup_text;
	}
	
	sprintf(msg, "%sE%d\n%s", MPI, (int)(txtlen + strlen(keystr)), keystr);
	hdrlen = strlen(msg);
	bcopy(text, msg + hdrlen, txtlen);
	
	msg[hdrlen + txtlen] = '\0';
	tcp_main_write(msg);
	sprintf (buf, "#completed session %s (%d)\n", sp->descr, sp->pid);
	write_message(buf);
	tcp_flush();
    } else {
	/* file wasn't changed, cancel editing session */
	sprintf(keystr, "C%d\n", sp->key);
	msg = (char*)malloc(64);
	if (!msg) {
	    errmsg("malloc");
	    goto cleanup_text;
	}
	sprintf(msg, "%sE%d\n%s", MPI, (int) strlen(keystr), keystr);
	
	tcp_main_write(msg);
	tcp_flush();
	sprintf(buf, "#cancelled session %s (%d)\n", sp->descr, sp->pid);
	write_message(buf);
    }
    
              free(msg);
cleanup_text: free(text);
cleanup_fd:   close(fd);
cleanup_file: if (unlink(sp->file) < 0)
		errmsg("unlink edit file");
}
